import logging
from concurrent.futures import Future

import minijson
from satella.coding import wraps, for_argument, silence_excs
from satella.coding.optionals import Optional
from satella.coding.predicates import x
from satella.coding.sequences import index_of
from satella.time import ExponentialBackoff

from ..orders import Order

import typing as tp
import select
from satella.coding.concurrent import TerminableThread

from ..exceptions import DataStreamSyncFailed, ConnectionFailed
from ..protocol import NGTTHeaderType
from .connection import NGTTSocket

logger = logging.getLogger(__name__)


def must_be_connected(fun):
    @wraps(fun)
    def outer(self, *args, **kwargs):
        if self.current_connection is None:
            self.connect()
            raise ConnectionFailed(True)
        return fun(self, *args, **kwargs)

    return outer


def encode_data(y) -> bytes:
    return minijson.dumps(y)


class NGTTConnection(TerminableThread):
    """
    A thread maintaining connection in the background.

    Note that instantiating this object is the same as calling start. You do not need to call
    start on this object after you initialize it.

    :ivar connected (bool) is connection opened
    """

    def __init__(self, cert_file: str, key_file: str,
                 on_new_order: tp.Callable[[Order], None]):
        super().__init__(name='ngtt uplink')
        self.on_new_order = on_new_order
        self.cert_file = cert_file
        self.stopped = False
        self.key_file = key_file
        self.current_connection = None
        self.currently_running_ops = []  # type: tp.List[tp.Tuple[NGTTHeaderType, bytes, Future]]
        self.op_id_to_op = {}  # type: tp.Dict[int, Future]
        logger.info('NGTT starting up')
        self.start()

    def stop(self, wait_for_completion: bool = True):
        """
        Stop this thread and the connection

        :param wait_for_completion: whether to wait for thread to terminate
        """
        if self.stopped:
            return
        self.terminate()
        if wait_for_completion:
            self.join()
        self.stopped = True

    def close(self):
        self.stop()
        if self.current_connection is not None:
            self.current_connection.close()
            self.current_connection = None
            self.op_id_to_op = {}

    @property
    @silence_excs(AttributeError, returns=False)
    def connected(self) -> bool:
        return self.current_connection.connected

    def connect(self):
        if self.connected:
            return
        eb = ExponentialBackoff(1, 30, self.safe_sleep)
        while not self.terminating and not self.connected:
            try:
                self.current_connection = NGTTSocket(self.cert_file, self.key_file)
                self.current_connection.connect()
            except ConnectionFailed as e:
                logger.warning('Failure reconnecting', exc_info=e)
                eb.failed()
                eb.sleep()

        if self.terminating:
            return

        self.op_id_to_op = {}
        for h_type, data, fut in self.currently_running_ops:
            id_ = self.current_connection.id_assigner.allocate_int()
            self.current_connection.send_frame(id_, h_type, data)
            self.op_id_to_op[id_] = fut
        logger.debug('Successfully connected')

    @must_be_connected
    @for_argument(None, encode_data)
    def sync_pathpoints(self, data) -> Future:
        """
        Try to synchronize pathpoints.

        This will survive multiple reconnection attempts.

        :param data: exactly the same thing that you would submit to POST
        at POST https://api.smok.co/v1/device/
        :return: a Future telling you whether this succeeds or fails
        """
        fut = Future()
        fut.set_running_or_notify_cancel()
        tid = self.current_connection.id_assigner.allocate_int()
        self.currently_running_ops.append((NGTTHeaderType.DATA_STREAM, data, fut))
        try:
            self.current_connection.send_frame(tid, NGTTHeaderType.DATA_STREAM, data)
        except ConnectionFailed:
            self.cleanup()
            raise
        self.op_id_to_op[tid] = fut
        return fut

    def inner_loop(self):
        logger.debug('Inner loop')
        self.current_connection.try_ping()
        ccon = [self.current_connection]
        rx, wx, ex = select.select(ccon,
                                   ccon if self.current_connection.wants_write else [], [],
                                   timeout=5)
        if not rx:
            return
        if wx:
            self.current_connection.try_send()
        frame = self.current_connection.recv_frame()
        if frame is None:
            logger.debug('Received nothing')
            return
        logger.debug('Received %s', frame)
        if frame.packet_type == NGTTHeaderType.PING:
            self.current_connection.got_ping()
        elif frame.packet_type == NGTTHeaderType.ORDER:
            try:
                data = minijson.loads(frame.data)
            except ValueError:
                logger.error('Received invalid JSON over the wire')
                raise ConnectionFailed('Got invalid JSON')
            order = Order(data, frame.tid, self.current_connection)
            self.on_new_order(order)
        elif frame.packet_type in (
                NGTTHeaderType.DATA_STREAM_REJECT, NGTTHeaderType.DATA_STREAM_CONFIRM):
            if frame.tid in self.op_id_to_op:
                # Assume it's a data stream running
                fut = self.op_id_to_op.pop(frame.tid)

                index = index_of(x[2] == fut, self.currently_running_ops)
                del self.currently_running_ops[index]

                if frame.packet_type == NGTTHeaderType.DATA_STREAM_CONFIRM:
                    fut.set_result(None)
                elif frame.packet_type == NGTTHeaderType.DATA_STREAM_REJECT:
                    fut.set_exception(DataStreamSyncFailed())
        elif frame.packet_type == NGTTHeaderType.SYNC_BAOB_RESPONSE:
            if frame.tid in self.op_id_to_op:
                fut = self.op_id_to_op.pop(frame.tid)

                index = index_of(x[2] == fut, self.currently_running_ops)
                del self.currently_running_ops[index]

                fut.set_result(frame.real_data)

    def loop(self) -> None:
        try:
            while not self.connected and not self.terminating:
                self.connect()
                logger.debug('Reconnecting')
            if self.terminating:
                return
            self.inner_loop()
        except ConnectionFailed as e:
            logger.debug('Connection failed, retrying', exc_info=e)
            self.cleanup()
            try:
                self.connect()
            except ConnectionFailed:
                pass

    def cleanup(self):
        Optional(self.current_connection).close()
        self.current_connection = None

    @must_be_connected
    @for_argument(None, encode_data)
    def sync_baobs(self, baobs) -> Future:
        """
        Request to synchronize BAOBs

        :param baobs: a dictionary of locally kept BAOB name => local version (tp.Dict[str, int])
        :return: a Future that will receive a result of dict
        {"download": [.. list of BAOBs to download from the server ..],
         "upload": [.. list of BAOBs to upload to the server ..]}

        :raises ConnectionFailed: connection failed
        """
        fut = Future()
        fut.set_running_or_notify_cancel()
        tid = self.current_connection.id_assigner.allocate_int()
        self.currently_running_ops.append((NGTTHeaderType.SYNC_BAOB_REQUEST, baobs, fut))
        try:
            self.current_connection.send_frame(tid, NGTTHeaderType.SYNC_BAOB_REQUEST, baobs)
        except ConnectionFailed:
            self.cleanup()
            raise
        self.op_id_to_op[tid] = fut
        return fut

    @must_be_connected
    @for_argument(None, encode_data)
    def stream_logs(self, data: tp.List) -> None:
        """
        Stream logs to the server

        This will work on a best-effort basis.

        :param data: the same thing that you would PUT /v1/device/device_logs
        """
        try:
            self.current_connection.send_frame(0, NGTTHeaderType.LOGS, data)
        except ConnectionFailed:
            self.cleanup()
            raise



try:
    import ujson as json
except ImportError:
    import json
import struct
import typing as tp

from satella.coding.structures import HashableIntEnum

from .exceptions import InvalidFrame


class NGTTHeaderType(HashableIntEnum):
    PING = 0  #: a PING, sent by the client, server responds alike
    ORDER = 1  #: sent by the server
    ORDER_CONFIRM = 2  #: sent by the client
    LOGS = 3  #: sent by the client
    DATA_STREAM = 4  #: sent by the client
    DATA_STREAM_CONFIRM = 5  #: sent by the server
    DATA_STREAM_REJECT = 6  #: sent by the server
    SYNC_BAOB_REQUEST = 7  #: sent by the client
    SYNC_BAOB_RESPONSE = 8  #: sent by the server


STRUCT_LHH = struct.Struct('>LHH')


class NGTTFrame:
    def __init__(self, tid: int, packet_type: NGTTHeaderType, data: bytes):
        self.tid = tid
        self.packet_type = packet_type
        self.data = data

    def __repr__(self) -> str:
        return f'NGTTFrame({self.tid}, {self.packet_type}, {self.data})'

    def __str__(self) -> str:
        return repr(self)

    @property
    def real_data(self) -> tp.Union[dict, list]:
        """
        :return: JSON unserialized data
        """
        return json.loads(self.data.decode('utf-8'))

    def __len__(self):
        return STRUCT_LHH.size + len(self.data)

    def __bytes__(self):
        return STRUCT_LHH.pack(len(self.data), self.tid, self.packet_type.value)

    @classmethod
    def from_bytes(cls, b: tp.Union[bytes, bytearray]) -> 'NGTTFrame':
        length, tid, h_type = STRUCT_LHH.unpack(b[:STRUCT_LHH.size])
        if h_type > NGTTHeaderType.DATA_STREAM_REJECT.value:
            raise InvalidFrame('Unrecognized packet type %s' % (h_type,))
        return NGTTFrame(tid, NGTTHeaderType(h_type), b[STRUCT_LHH.size:STRUCT_LHH.size + length])


def env_to_hostname(env: int) -> str:
    return {0: 'api.smok.co',
            1: 'api.test.smok-serwis.pl'}.get(env, 'rapid-rs')



import typing as tp

from ngtt.protocol import NGTTHeaderType


class Order:
    """
    An order received from the server.

    Can be confirmed via :meth:`~ngtt.orders.Order.acknowledge` to signal to the server
    that it's been processed.
    """
    __slots__ = ('data', 'tid', 'sock', 'confirmed')

    def __init__(self, data: tp.Dict, tid: int, sock: 'NGTTSocket'):
        self.data = data
        self.tid = tid
        self.sock = sock
        self.confirmed = False

    def acknowledge(self):
        """
        Signal to the server that the order has been processed
        """
        if not self.confirmed:
            self.sock.send_frame(self.tid, NGTTHeaderType.ORDER_CONFIRM, b'')
            self.confirmed = True



class NGTTError(Exception):
    pass


class ConnectionFailed(NGTTError):
    """
    :ivar is_due_to_no_internet: was this error caused by the internet not being there?
    """
    def __init__(self, is_due_to_no_internet: bool = False):
        self.is_due_to_no_internet = is_due_to_no_internet


class DataStreamSyncFailed(NGTTError):
    pass


class InvalidFrame(NGTTError):
    pass



import typing as tp

from ngtt.exceptions import ConnectionFailed, DataStreamSyncFailed
from ngtt.orders import Order
from smok.pathpoint.orders import sections_from_list
from smok.sync_workers.base import BaseSyncWorker, SyncError
from ngtt.uplink import NGTTConnection


class NGTTSyncWorker(BaseSyncWorker):

    def sync_logs(self, data: tp.List[dict]):
        try:
            self.connection.stream_logs(data)
        except ConnectionFailed as e:
            raise SyncError(e.is_due_to_no_internet)

    def __init__(self, device: 'SMOKDevice'):
        super().__init__(device, True)
        self.connection = NGTTConnection(device.temp_file_for_cert,
                                         device.temp_file_for_key,
                                         self.process_orders)

    def process_orders(self, orders: Order):
        sections = sections_from_list(orders.data)
        sections[-1].future.add_done_callback(lambda fut: orders.acknowledge())
        for sec in sections:
            self.device.executor.queue.put(sec)

    def sync_pathpoints(self, data: tp.List[dict]):
        """
        :raises SyncError: on failure
        """
        try:
            self.connection.sync_pathpoints(data).result()
        except DataStreamSyncFailed:
            raise SyncError(False, True)

    def close(self):
        self.connection.close()



import logging
import os
import socket
import ssl
import tempfile
import threading
import time
import typing as tp
from ssl import SSLContext, PROTOCOL_TLS_CLIENT, SSLError, CERT_REQUIRED

from satella.coding import silence_excs, reraise_as, Closeable, wraps
from satella.coding.concurrent import IDAllocator
from satella.coding.optionals import Optional
from satella.files import read_in_file
from satella.instrumentation import Traceback

from smok.client import get_dev_ca_cert, get_root_cert
from smok.client.certificate import get_device_info
from ..exceptions import ConnectionFailed
from ..protocol import NGTTHeaderType, STRUCT_LHH, env_to_hostname, NGTTFrame

PING_INTERVAL_TIME = 30
logger = logging.getLogger(__name__)


def must_be_connected(fun):
    @wraps(fun)
    def outer(self, *args, **kwargs):
        if not self.connected:
            self.connect()
        return fun(self, *args, **kwargs)

    return outer


class NGTTSocket(Closeable):
    @property
    def wants_write(self) -> bool:
        return bool(self.w_buffer)

    def __init__(self, cert_file: str, key_file: str):
        logger.info('New connection %s %s', cert_file, key_file)
        self.socket = None
        self.connected = False
        self.connection_lock = threading.Lock()
        environment = get_device_info(read_in_file(cert_file))[1]
        self.host = env_to_hostname(environment)
        logger.info('Environment is %s', environment)
        self.cert_file = cert_file
        self.key_file = key_file
        self.buffer = bytearray()
        self.w_buffer = bytearray()
        self.ping_id = None
        self.last_read = None
        try:
            with tempfile.NamedTemporaryFile('wb', delete=False) as chain_file:
                chain_file.write(get_dev_ca_cert())
                chain_file.write(b'\n')
                chain_file.write(get_root_ca_cert())
                chain_file.close()
                self.chain_file_name = chain_file.name
        except Exception as e:
            logger.error('Found exception %s', e, exc_info=e)
            raise

        self.id_assigner = IDAllocator(start_at=1)
        super().__init__()

    @reraise_as(ssl.SSLError, ConnectionFailed)
    @silence_excs(ssl.SSLWantWriteError)
    @must_be_connected
    def send_frame(self, tid: int, header: NGTTHeaderType, data: bytes = b'') -> None:
        """
        Schedule a frame to be sent

        :param tid: transaction ID
        :param header: packet type
        :param data: data to send
        """
        if self.closed:
            return
        logger.debug('Sending %s', NGTTFrame(tid, header, data))
        self.w_buffer.extend(STRUCT_LHH.pack(len(data), tid, header.value))
        self.w_buffer.extend(data)
        data_sent = self.socket.send(self.w_buffer)
        del self.w_buffer[:data_sent]

    @reraise_as(ssl.SSLError, ConnectionFailed)
    @silence_excs(ssl.SSLWantWriteError)
    @must_be_connected
    def try_send(self):
        """
        Try to send some data
        """
        if self.w_buffer:
            try:
                data_sent = self.socket.send(self.w_buffer)
                del self.w_buffer[:data_sent]
            except socket.timeout:
                return

    @must_be_connected
    def try_ping(self):
        if time.monotonic() - self.last_read > PING_INTERVAL_TIME and self.ping_id is None:
            self.ping_id = self.id_assigner.allocate_int()
            self.send_frame(self.ping_id, NGTTHeaderType.PING, b'')

    @must_be_connected
    def got_ping(self):
        if self.ping_id is not None:
            self.id_assigner.mark_as_free(self.ping_id)
            self.ping_id = None

    def fileno(self) -> int:
        return self.socket.fileno()

    @reraise_as(ssl.SSLError, ConnectionFailed)
    @silence_excs(ssl.SSLWantReadError)
    @must_be_connected
    def recv_frame(self) -> tp.Optional[NGTTFrame]:
        """
        Receive a frame from remote socket, or None if nothing could be assembled as of now.

        :raises ConnectionFailed: connection closed
        :return: a tuple of transaction ID, header type, data
        """
        data = self.socket.recv(512)
        if not data:
            raise ConnectionFailed()
        self.last_read = time.monotonic()
        self.buffer.extend(data)
        if len(self.buffer) > STRUCT_LHH.size:
            length, tid, h_type = STRUCT_LHH.unpack(self.buffer[:STRUCT_LHH.size])
            if len(self.buffer) < STRUCT_LHH.size + length:
                return
            data = self.buffer[STRUCT_LHH.size:STRUCT_LHH.size + length]
            del self.buffer[:STRUCT_LHH.size + length]
            return NGTTFrame(tid, NGTTHeaderType(h_type), data)

    def close(self, wait_for_me: bool = True):
        logger.info('Closing %s %s %s', self.closed, self.connected, self.socket)
        if super().close():
            logger.info('Actually closing')
            self.disconnect()
            try:
                os.unlink(self.chain_file_name)
            except TypeError:
                logger.error('Tried to close an uninitialized object')
                logger.warning(Traceback().pretty_print())
            except OSError as e:
                logger.error('Failure to remove certificate chain file %s', self.chain_file_name,
                             exc_info=e)

    def disconnect(self):
        """
        Disconnect from the remote host
        """
        if self.socket is not None:
            self.socket.close()
            self.socket = None
            self.connected = False

    def connect(self):
        """
        Connect to remote host

        :raises SSLError: an error occurred
        :raises RuntimeError: upon connection being closed
        """
        with self.connection_lock:
            if self.closed:
                raise RuntimeError('This connection is closed!')
            if self.connected:
                return
            ssl_context = SSLContext(PROTOCOL_TLS_CLIENT)
            ssl_context.load_verify_locations(self.chain_file_name)
            ssl_context.load_cert_chain(self.cert_file, self.key_file)
            ssl_context.verify_mode = CERT_REQUIRED
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(10)
            ssl_sock = ssl_context.wrap_socket(sock, server_hostname=self.host)
            try:
                ssl_sock.connect((self.host, 2408))
                ssl_sock.do_handshake()
            except (socket.error, SSLError) as e:
                logger.error(Traceback().pretty_print())
                ssl_sock.close()
                raise ConnectionFailed(True) from e
            self.socket = ssl_sock
            self.socket.setblocking(False)
            self.last_read = time.monotonic()
            self.buffer = bytearray()
            self.w_buffer = bytearray()
            self.connected = True
